/*
 * Look for TODO comments in this file for suggestions on how to implement
 * your parser.
 */
package parser;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;
import java.util.Queue;
import java.util.Set;
import java.util.Stack;
import lexer.ExprLexer;
import lexer.ParenLexer;
import lexer.TinyLexer;
import org.antlr.v4.runtime.ANTLRFileStream;
import org.antlr.v4.runtime.Lexer;
import org.antlr.v4.runtime.Token;
import org.antlr.v4.runtime.Vocabulary;
import parser.util.SymbolComparator;

/**
 *
 * @author edwajohn
 */
public class Parser {

  final Grammar grammar;

  /**
   * Action table for bottom-up parsing. Accessed as
   * actionTable.get(state).get(terminal).
   */
  private final HashMap<Integer, HashMap<String, Action>> actionTable;
  /**
   * Goto table for bottom-up parsing. Accessed as gotoTable.get(state).get(NT).
   */
  private final HashMap<Integer, HashMap<String, Integer>> gotoTable;

  public Parser(String grammarFilename) throws IOException {
    grammar = new Grammar(grammarFilename);

    // TODO: construct the action and goto tables. Note, to choose which
    // lookahead symbols to use for the Items (see figure 3.19), use the
    // FOLLOW set of the non-terminal on the left-hand side of the rule.
    // actionTable = new HashMap<>();
//    gotoTable = new HashMap<>();

    // Compute all possible items
    ArrayList<Item> items = computeAllPossibleItems(grammar.rules, grammar.follow);

    System.out.println("Items:");
    for (Item item : items) {
      System.out.println(item.toString());
    }

    // Compute the closure of the first item
//    Set<Set<Item>> CC = new HashSet<>();
    HashMap<Set<Item>, Integer> state2name = new HashMap<>();
    HashMap<Integer, Set<Item>> name2state = new HashMap<>();

    gotoTable = computeGotoTable(grammar, items, state2name, name2state);
    actionTable = computeActionTable(
            state2name, name2state, gotoTable, grammar.startSymbol);

  }

  // TODO: Implement this method. You'll iterate through all rules, look at all
  // possible lookaheads (anything that follows the lhs -- use the follow set)
  // and then create the item using Item.create().
  static public ArrayList<Item> computeAllPossibleItems(List<Rule> rules, HashMap<String, HashSet<String>> follow) {
    ArrayList<Item> items = new ArrayList<>();
    for (Rule rule : rules) {
      for (String lookahead : follow.get(rule.getLhs())) {
        items.addAll(Item.create(rule, lookahead));
      }
    }
    return items;
  }

  // Pass-through function to computeClosure implementation.
  static public Set<Item> computeClosure(
          Item seed, ArrayList<Item> items, Grammar grammar) {
    Set<Item> seeds = new HashSet<>();
    seeds.add(seed);
    return computeClosure(seeds, items, grammar);
  }

  // TODO: implement computeClosure.
  // You'll get the nonterminals using grammar.nonterminals.
  static public Set<Item> computeClosure(
          Set<Item> seed, ArrayList<Item> items, Grammar grammar) {
    HashMap<String, HashSet<String>> first = grammar.first;
    boolean changed = true;
    Set<Item> closure = new HashSet<>();
    closure.addAll(seed);
    while (changed) {
      changed = false;
      Set<Item> toAdd = new HashSet<>();
      for (Item item : closure) {
        String next = item.getNextSymbol();
        String delta = item.getNextNextSymbol();
        if (grammar.nonterminals.contains(next)) {
          for (Rule rule : grammar.nt2rules.get(next)) {
            boolean addLookahead = (delta == null);
            if (delta != null) {
              for (String b : first.get(delta)) {
                toAdd.add(new Item(rule, 0, b));
                if (b.equals(Util.EPSILON)) {
                  addLookahead = true;
                }
              }
            }
            if (addLookahead) {
              toAdd.add(new Item(rule, 0, item.getA()));
            }
          }
        }
      }
      final int size = closure.size();
      closure.addAll(toAdd);
      changed = closure.size() > size;
    }

    return closure;
  }

  public static Set<Item> computeGoto(
          Set<Item> state, String x, ArrayList<Item> items, Grammar grammar) {
    Set<Item> moved = new HashSet<>();
    for (Item item : state) {
      if (x.equals(item.getNextSymbol())) {
        moved.add(item.advance());
      }
    }
    return computeClosure(moved, items, grammar);
  }

  // input: grammar, items
  // output: state2name, name2state
  public static HashMap<Integer, HashMap<String, Integer>> computeGotoTable(
          Grammar grammar,
          ArrayList<Item> items,
          HashMap<Set<Item>, Integer> state2name,
          HashMap<Integer, Set<Item>> name2state) {
    // Fill the goto table
    HashMap<Integer, HashMap<String, Integer>> gotoTable = new HashMap<>();

    Set<Item> CC0 = computeClosure(items.get(0), items, grammar);
    // System.out.println("Closure of CC0");
    // for (Item item : CC0) {
    //   System.out.println(item.toString());
    // }

    int nextName = 0;
    final int CC0name = nextName++;
    state2name.put(CC0, CC0name);
    name2state.put(CC0name, CC0);
    gotoTable.put(CC0name, new HashMap<>());
    Queue<Set<Item>> queue = new LinkedList<>();
    queue.add(CC0);
    while (!queue.isEmpty()) {
      Set<Item> CCi = queue.remove();
      Set<String> lookaheads = new HashSet<String>();
      for (Item item : CCi) {
        String x = item.getNextSymbol();
        if (x != null) {
          lookaheads.add(x);
        }
      }
      for (String x : lookaheads) {
        Set<Item> CCj = computeGoto(CCi, x, items, grammar);
//        if (!CC.contains(CCj)) {
        if (!state2name.containsKey(CCj)) {
          // First time we've encountered this state.
          final int name = nextName++;
          state2name.put(CCj, name);
          name2state.put(name, CCj);
          if (!gotoTable.containsKey(name)) {
            gotoTable.put(name, new HashMap<>());
          }
//          CC.add(CCj);
          queue.add(CCj);
        }
        gotoTable.get(state2name.get(CCi)).put(x, state2name.get(CCj));
      }
    }
    return gotoTable;
  }

  // input: state2name, name2state, gotoTable
  public static HashMap<Integer, HashMap<String, Action>> computeActionTable(
          HashMap<Set<Item>, Integer> state2name,
          HashMap<Integer, Set<Item>> name2state,
          HashMap<Integer, HashMap<String, Integer>> gotoTable,
          String startSymbol) {
    // TODO: Fill the action table. Note that if the item is of the form
    // of a reduce, then a can be EOF.

    HashMap<Integer, HashMap<String, Action>> actionTable = new HashMap<>();

    for (Integer CCiName : name2state.keySet()) {
      Set<Item> CCi = name2state.get(CCiName);
      if (!actionTable.containsKey(CCiName)) {
        actionTable.put(CCiName, new HashMap<>());
      }
      for (Item I : CCi) {
        final String c = I.getNextSymbol();
        final String a = I.getA();
        if (c != null && gotoTable.containsKey(CCiName) && gotoTable.get(CCiName).containsKey(c)) {
          final Integer CCjName = gotoTable.get(CCiName).get(c);
          actionTable.get(CCiName).put(c, Action.createShift(CCjName));
        } else if (I.getRule().getLhs().equals(startSymbol) && c == null && a.equals(Util.EOF)) {
          actionTable.get(CCiName).put(a, Action.createAccept());
        } else if (c == null) {
          actionTable.get(CCiName).put(a, Action.createReduce(I.getRule()));
        }
      }
    }
    return actionTable;
  }

  private Lexer scanFile(String filename) throws IOException {
    // We use ANTLR's scanner (lexer) to produce the tokens.
    System.out.println("\nReading input file " + filename + "\n");
    Lexer scanner = null;
    final ANTLRFileStream fs = new ANTLRFileStream(filename);
    switch (grammar.grammarName) {
      case "Paren":
        scanner = new ParenLexer(fs);
        break;
      case "Expr":
        scanner = new ExprLexer(fs);
        break;
      case "Tiny":
        scanner = new TinyLexer(fs);
        break;
      default:
        System.out.println("Unknown scanner");
        break;
    }

    return scanner;
  }

  /**
   * Parses an input file. On an error, throw a ParserException using the
   * factory method ParserException.create().
   *
   * @param filename
   * @throws IOException
   */
  public void parseBottomUp(String filename) throws IOException, ParserException {
  }



}
